/*
  Simple Demo for GLSL
  www.lighthouse3d.com
  
  tweaked a bit more by andy
  
  0 / 1 / 2 / 3 / 4 / 5 change the convolution kernel
  a / z / s change the image
*/

#include "GL/glew.h"

#ifdef __linux__
#include <GL/glut.h>
#endif

#ifdef __APPLE__
#include <glut.h>
#endif

#ifdef _WIN32
#include "glut.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

int whichMode = 0;
float amount = 1.0;
int winWidth = 512;
int winHeight = 512;


char texData0[800000];
char texData1[800000];
char frameBuf[800000];
char frameBufFlip[800000];

float mouseWindowX = 0.0;
float mouseWindowY = 0.0;

GLuint tex0, tex1, chosenTex, frame, frameFlip;

float rot;

////////////////////////////////////////////////////////////////////////////////

char *textFileRead(char *fn)
{
	FILE *fp;
	char *content = NULL;

	int count=0;

	if (fn != NULL) {
		fp = fopen(fn,"rt");

		if (fp != NULL) {
      
      fseek(fp, 0, SEEK_END);
      count = ftell(fp);
      rewind(fp);

			if (count > 0) {
				content = (char *)malloc(sizeof(char) * (count+1));
				count = fread(content,sizeof(char),count,fp);
				content[count] = '\0';
			}
			fclose(fp);
		}
	}
	
	if (content == NULL)
	   {
	   fprintf(stderr, "ERROR: could not load in file %s\n", fn);
	   exit(1);
	   }
	return content;
}                    


////////////////////////////////////////////////////////////////////////////////

GLuint p, p2;

float lpos[4] = {1.0, 0.5, 1.0, 0.0};

// for now keep the action in the lower left 512 by 512 of the window
// so the texture sizes are OK
void changeSize(int w, int h) {

	// Prevent a divide by zero, when window is too short
	// (you cant make a window of zero width).
	
	if(h == 0)
		h = 1;

	float ratio = 1.0* winWidth / winHeight;

	// Reset the coordinate system before modifying
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	winWidth = winWidth;
	winHeight = winWidth;
	
	// Set the viewport to be the entire window
    glViewport(0, 0, winWidth, winHeight);

	// Set the correct perspective.
	gluPerspective(45, ratio, 1, 1000);
	glMatrixMode(GL_MODELVIEW);
}

////////////////////////////////////////////////////////////////

void PassiveMouseMotion(int x, int y) {   
	if (x < 0)
		mouseWindowX = -1;
	else if (x > winWidth)
		mouseWindowX = 1;
	else
		mouseWindowX = 2 * (x / (float) winWidth) - 1.0;
           
	if (y < 0)
		mouseWindowY = -1;
	else if (y > winHeight)
		mouseWindowY = 1;
	else
		mouseWindowY = 2 * (y / (float) winHeight) - 1.0;
	}
	
////////////////////////////////////////////////////////////////////////////////

void printShaderLog(GLuint prog)
{
    GLint infoLogLength = 0;
    GLsizei charsWritten  = 0;
    GLchar *infoLog;

    glGetShaderiv(prog, GL_INFO_LOG_LENGTH, &infoLogLength);

    if (infoLogLength > 0)
    {
        infoLog = (char *) malloc(infoLogLength);
        glGetShaderInfoLog(prog, infoLogLength, &charsWritten, infoLog);
		printf("%s\n",infoLog);
        free(infoLog);
    }
}

void printProgramLog(GLuint shad)
{
    GLint infoLogLength = 0;
    GLsizei charsWritten  = 0;
    GLchar *infoLog;

    glGetProgramiv(shad, GL_INFO_LOG_LENGTH, &infoLogLength);

    if (infoLogLength > 0)
    {
        infoLog = (char *) malloc(infoLogLength);
        glGetProgramInfoLog(shad, infoLogLength, &charsWritten, infoLog);
		printf("%s\n",infoLog);
        free(infoLog);
    }
}


////////////////////////////////////////////////////////////////////////////////

float rotAmount = 0;

void renderScene(void) {

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	gluLookAt(0.0,0.0,5.0, 
		      0.0,0.0,-1.0,
			  0.0f,1.0f,0.0f);
	
	glDisable(GL_TEXTURE_2D);
	glUseProgram(0);


	glColor3f(1.0, 1.0, 1.0);	
	
	glPushMatrix();
		glRotatef(mouseWindowX*90+90,0,1,0); // rotate the teapot
		glRotatef(mouseWindowY*90,0,0,1);
	
		glutSolidTeapot(1.0);
	glPopMatrix();
	
	rotAmount += 0.1;

	// now read the pixels back out of the frame buffer
    glBindTexture(GL_TEXTURE_2D, frame);
    glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, winWidth, winHeight);
        
    int size, end;
    int h, w, rowSize;

    size = 3*sizeof(char);
    end = winWidth*winHeight*size;
    rowSize = winWidth*size;
   
    // read the texture back out of the buffer here ///
    glGetTexImage(GL_TEXTURE_2D, 0, GL_RGB, GL_UNSIGNED_BYTE, frameBuf);
              
    // want to do a vertical flip on the texture
    	for (h=0; h<winHeight;h++)
    		memcpy(&frameBufFlip[end-h*rowSize],
    			   &frameBuf[h*rowSize], rowSize);
 
 	// store that flipped texture and then use it
    glBindTexture(GL_TEXTURE_2D, frameFlip);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 512, 512, 0, GL_RGB,
        GL_UNSIGNED_BYTE, (const GLvoid *) frameBufFlip);
        
    // need to clear the teapot out of the buffer    
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


    glEnable(GL_TEXTURE_2D); 

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, chosenTex);
    
    glUseProgram(p);

    GLint texLoc, tex2Loc, alphaLoc, whichPtr;
    static float blender = 0.0;

	// always using texture 0
    texLoc = glGetUniformLocation(p, "colorMap");
    glUniform1i(texLoc, 0);

    whichPtr = glGetUniformLocation(p, "whichOne");
    glUniform1i(whichPtr, whichMode);

            
    glColor3f(1.0, 1.0, 1.0);
    glBegin( GL_POLYGON );
        glTexCoord2f( 0.0f, 1.0f );
        glVertex3f( -2.0f, -2.0f, 0.0f );
        
        glTexCoord2f( 1.0f, 1.0f ); 

        glVertex3f(  2.0f, -2.0f, 0.0f );
        
        glTexCoord2f( 1.0f, 0.0f );

        glVertex3f(  2.0f,  2.0f, 0.0f );
        
        glTexCoord2f( 0.0f, 0.0f ); 
        glVertex3f( -2.0f,  2.0f, 0.0f );
   glEnd();

	glutSwapBuffers();
}

////////////////////////////////////////////////////////////////////////////////

void processNormalKeys(unsigned char key, int x, int y) {

	if (key == 27) 
		exit(0);

    if (key == '0') 
		{
		whichMode = 0;
		amount = 1.0;
		}
    if (key == '1') 
		{
		whichMode = 1;
		amount = 1.0;
		}
    if (key == '2') 
		{
		whichMode = 2;
		amount = 1.0;
		}
    if (key == '3') 
		{
		whichMode = 3;
		amount = 1.0;
		}
    if (key == '4') 
		{
		whichMode = 4;
		amount = 1.0;
		}
    if (key == '5') 
		{
		whichMode = 5;
		amount = 1.0;
		}
		
	if ((key == 'A') || (key == 'a'))
		{
		chosenTex = tex0;
		}
	if ((key == 'Z') || (key == 'z'))
		{
		chosenTex = tex1;
		}
	if ((key == 'S') || (key == 's'))
		{
		chosenTex = frameFlip;
		}
}

////////////////////////////////////////////////////////////////////////////////

GLuint setShaders(char * vert, char * frag) {
    GLuint v,f, pro;
	char *vs,*fs;

	v = glCreateShader(GL_VERTEX_SHADER);
	f = glCreateShader(GL_FRAGMENT_SHADER);

	vs = textFileRead(vert);
	fs = textFileRead(frag);

	const char * vv = vs;
	const char * ff = fs;

	glShaderSource(v, 1, &vv,NULL);
	glShaderSource(f, 1, &ff,NULL);

	free(vs);free(fs);

	glCompileShader(v);
	glCompileShader(f);

	printShaderLog(v);
	printShaderLog(f);

	pro = glCreateProgram();
	glAttachShader(pro,v);
	glAttachShader(pro,f);

	glLinkProgram(pro);
	printProgramLog(pro);
	
	return(pro);
}

////////////////////////////////////////////////////////////////////////////////

GLuint setupTexture(char * texData)
{
	GLuint textureNum[1];

	glGenTextures(1, textureNum);

    glBindTexture(GL_TEXTURE_2D, textureNum[0]);

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 512, 512, 0, GL_RGB,
        GL_UNSIGNED_BYTE, (const GLvoid *) texData);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	return(textureNum[0]);
}

////////////////////////////////////////////////////////////////////////////////

void doTexStuff(char * dataPath, char * fileName, int width, int height,
		char * TheArray)
    {
    char fullPath[256];
    FILE * imageFile;

    strcpy(fullPath, dataPath);
    strcat(fullPath, fileName);
    imageFile = fopen(fullPath, "r");
	
    if (imageFile == NULL) 
	   fprintf(stderr,"Cannot find texture file in data directory\n");
    else
        {
        fread(TheArray, sizeof(char), height*width*3, imageFile);
        fclose(imageFile);
        }
    }

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutInitWindowPosition(100,100);
	glutInitWindowSize(winWidth, winHeight);
	glutCreateWindow("GLSL");

	glutDisplayFunc(renderScene);
	glutIdleFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutKeyboardFunc(processNormalKeys);
	glutPassiveMotionFunc(PassiveMouseMotion);

	glewInit();
	
	glEnable(GL_DEPTH_TEST);
	glClearColor(0.0,0.0,0.0,1.0);

    doTexStuff((char *) "./", (char *) "goog.raw",  512, 512, texData0);
    tex0 = setupTexture(texData0);
    doTexStuff((char *) "./", (char *) "sq.raw",  512, 512, texData1);
    tex1 = setupTexture(texData1);
    
    frame = setupTexture(frameBuf);
    frameFlip = setupTexture(frameBufFlip);
    
    chosenTex = tex1;

	if (GLEW_ARB_vertex_shader && GLEW_ARB_fragment_shader)
		printf("Ready for GLSL\n");
	else {
		printf("No GLSL support\n");
		exit(1);
	}

	p = setShaders((char *) "./convolve.vert", (char *) "./convolve.frag");
	
	glutMainLoop();
	return 0;
}

